#include <asm/types.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <iostream>
#include <linux/videodev2.h>
#include <malloc.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "hdmi_capture_util.h"
#include "multi_channel_camera.hpp"

#define CLEAR(x) memset(&(x), 0, sizeof(x))

int errno_exit(const char *s) {
    fprintf(stderr, "%s error %d, %s\n", s, errno, strerror(errno));
    return -1;
}

int xioctl(int fh, int request, void *arg) {
    int r = 0;
    do {
        r = ioctl(fh, request, arg);
    } while (-1 == r && EINTR == errno);
    return r;
}

int MultiChannelCamera::process_image(void *src, int size) {
    int ret = 0;
    struct frame_info_t frame_info;
    ret = decode_frame_header((const uint8_t *)src, size, &frame_info);
    if (ret != 0) {
        printf("decode_frame_info failed. [%d]\n", ret);
        return -1;
    }
    // For debug
    // if (log_en)
    //     printf("[MultiChannelCameraLog]:ch_num [%d] frame_num[%d].\n",
    //            frame_info.ch_num, frame_info.frame_num);

    // For debug
    {
        static uint32_t frame_num[16] = {0};
        static uint8_t first_flag[16] = {0};
        if (first_flag[frame_info.ch_num] == 1) {
            if (frame_num[frame_info.ch_num] + 1 != frame_info.frame_num) {
                printf("Warning!!!,[MultiChannelCameraLog] Frame Num "
                       "Nonsequence: ch_num "
                       "[%d] last "
                       "frame_num[%d]-frame_num[%d].\n",
                       frame_info.ch_num, frame_num[frame_info.ch_num],
                       frame_info.frame_num);
            }
            frame_num[frame_info.ch_num] = frame_info.frame_num;
        } else {
            first_flag[frame_info.ch_num] = 1;
            frame_num[frame_info.ch_num] = frame_info.frame_num;
        }
    }
    if (frame_info.type == VIDEO_FRAME_NORMAL) {
        if (_cb_read != nullptr) {
            _cb_read(frame_info.ch_num, frame_info.frame_num, src, size, _width,
                     _height, _user_data);
        }

    } else {
        /*  注意：因为 cache的原因，对于空帧 也需要内存拷贝
            否则回影响后面空帧的数据接收导致 空帧的帧头的解析。
        */
        memcpy(_empty_save_buf, src,
               (size > _buffers_size) ? _buffers_size : size);
        printf("Warning!!!,[MultiChannelCameraLog]:Recv Null Frame : ch_num "
               "[%d] frame_num[%d].\n",
               frame_info.ch_num, frame_info.frame_num);
    }

    return 0;
}

int MultiChannelCamera::read_frame() {
    struct v4l2_buffer buf;
    CLEAR(buf);
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_USERPTR;

    if (0 != xioctl(_fd, VIDIOC_DQBUF, &buf)) {
        switch (errno) {
        case EAGAIN:
            return 0;
        case EIO:
            /* Could ignore EIO, see spec. */
            /* fall through */
        default:
            return errno_exit("VIDIOC_DQBUF");
        }
    }
    /*
     * https://www.kernel.org/doc/html/v4.9/media/uapi/v4l/vidioc-qbuf.html
     * 注意：
     *   因为 VIDIOC_QBUF 接口返回的数据是 正在拷贝的bufer,
     * 所以此处使用上次的获得的bufer 或者延迟 一帧的时间再拷贝数据
     *
     *   其他摄像头程序是不需要这些操作的：上一帧和本帧基本无差别
     *   本程序的适用情况： 多个通道的视频公用同一个 通道
     */
    if (_last_buf.m.userptr != 0) {
        // invalidate_device_memeory(buffers[_last_buf.index].start,_last_buf.bytesused);
        process_image(_pbuffers[_last_buf.index].start, _last_buf.bytesused);
        // memset(buffers[_last_buf.index].start,0,_last_buf.bytesused);
        if (-1 == xioctl(_fd, VIDIOC_QBUF, &_last_buf)) {
            return errno_exit("VIDIOC_QBUF");
        }
    }

    _last_buf = buf;
    return 0;
}

void MultiChannelCamera::run() {
    int r = 0;
    while (_loop) {
        fd_set fds;
        struct timeval tv;

        FD_ZERO(&fds);
        FD_SET(_fd, &fds);

        tv.tv_sec = 2;
        tv.tv_usec = 0;

        r = select(_fd + 1, &fds, NULL, NULL, &tv);
        if (-1 == r) {
            if (EINTR == errno) {
                continue;
            }
            errno_exit("select");
            break;
        }
        if (0 == r) {
            fprintf(stderr, "select timeout,wait %d s\n", tv.tv_sec);
            continue;
        }
        if (read_frame()) {
            fprintf(stderr, "camera read loop break, read_frame error.\n");
            break;
        }
    }
}

int MultiChannelCamera::open_device(void) {
    struct stat st;
    if (-1 == stat(_dev_name.c_str(), &st)) {
        fprintf(stderr, "Cannot identify '%s': %d, %s\n", _dev_name, errno,
                strerror(errno));
        return -1;
    }

    if (!S_ISCHR(st.st_mode)) {
        fprintf(stderr, "%s is not a char device", _dev_name.c_str());
        return -1;
    }

    _fd = open(_dev_name.c_str(), O_RDWR /* required */ | O_NONBLOCK, 0);
    if (-1 == _fd) {
        fprintf(stderr, "Cannot open '%s': %d, %s\n", _dev_name, errno,
                strerror(errno));
        return -1;
    }
    return 0;
}

int MultiChannelCamera::init_userp(uint32_t buffer_size) {

    struct v4l2_requestbuffers req;
    CLEAR(req);
    req.count = _buffers_count;
    req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory = V4L2_MEMORY_USERPTR;

    if (-1 == xioctl(_fd, VIDIOC_REQBUFS, &req)) {
        if (EINVAL == errno) {
            fprintf(stderr, "%s does not support user pointer i/o/n",
                    _dev_name.c_str());
            return -1;
        } else {
            fprintf(stderr, "%s ioctl error: VIDIOC_REQBUFS\\n",
                    _dev_name.c_str());
            return -1;
        }
    }

    int ret = init_device_memory();
    if (ret < 0) {
        fprintf(stderr, "%s init_device_memory ...\\n", _dev_name.c_str());
        return -1;
    }

    _pbuffers =
        (struct buffer_t *)calloc(_buffers_count, sizeof(struct buffer_t));
    if (!_pbuffers) {
        fprintf(stderr, "Out of memory/n");
        return -1;
    }

    for (int i = 0; i < _buffers_count; i++) {
        _pbuffers[i].length = buffer_size;
        _pbuffers[i].start = (void *)malloc_device_memory(buffer_size);
        if (!_pbuffers[i].start) {
            fprintf(stderr, "Out of memory/n");
            return -1;
        }
    }
    return 0;
}

int MultiChannelCamera::init_device(void) {
    struct v4l2_capability cap;
    struct v4l2_format fmt;
    unsigned int min = 0;

    if (-1 == xioctl(_fd, VIDIOC_QUERYCAP, &cap)) {
        if (EINVAL == errno) {
            fprintf(stderr, "%s is no V4L2 device\\n", _dev_name.c_str());
            return -1;
        } else {
            fprintf(stderr, "%s ioctl error: VIDIOC_QUERYCAP\\n",
                    _dev_name.c_str());
            return -1;
        }
    }

    if (!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)) {
        fprintf(stderr, "%s is no video capture device\\n", _dev_name.c_str());
        return -1;
    }

    if (!(cap.capabilities & V4L2_CAP_STREAMING)) {
        fprintf(stderr, "%s does not support streaming i/o\\n",
                _dev_name.c_str());
        return -1;
    }
    CLEAR(fmt);
    fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width = _width;
    fmt.fmt.pix.height = _height;
    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24;
    fmt.fmt.pix.field = V4L2_FIELD_NONE;

    if (-1 == xioctl(_fd, VIDIOC_S_FMT, &fmt)) {
        fprintf(stderr, "%s ioctl error: VIDIOC_S_FMT\\n", _dev_name.c_str());
        return -1;
    }

    /* Buggy driver paranoia. */
    min = fmt.fmt.pix.width * 3;
    if (fmt.fmt.pix.bytesperline < min) {
        fmt.fmt.pix.bytesperline = min;
    }
    min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
    if (fmt.fmt.pix.sizeimage < min) {
        fmt.fmt.pix.sizeimage = min;
    }
    _buffers_size = fmt.fmt.pix.sizeimage;
    return init_userp(_buffers_size);
}

int MultiChannelCamera::start_capturing(void) {
    unsigned int i = 0;
    enum v4l2_buf_type type;

    for (i = 0; i < _buffers_count; ++i) {
        struct v4l2_buffer buf;

        CLEAR(buf);

        buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory = V4L2_MEMORY_USERPTR;
        buf.index = i;
        buf.m.userptr = (unsigned long)_pbuffers[i].start;
        buf.length = _pbuffers[i].length;

        if (-1 == xioctl(_fd, VIDIOC_QBUF, &buf)) {
            fprintf(stderr, "%s ioctl error: VIDIOC_QBUF\\n",
                    _dev_name.c_str());
            return -1;
        }
    }

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (-1 == xioctl(_fd, VIDIOC_STREAMON, &type)) {
        fprintf(stderr, "%s ioctl error: VIDIOC_STREAMON\\n",
                _dev_name.c_str());
        return -1;
    }
    return 0;
}

int MultiChannelCamera::start() {
    int ret = 0;
    ret = open_device();
    if (ret != 0) {
        return ret;
    }
    ret = init_device();
    if (ret != 0) {
        return ret;
    }
    start_capturing();
    if (ret != 0) {
        return ret;
    }
    _empty_save_buf = (uint8_t *)malloc(_buffers_size);
    if (_empty_save_buf == nullptr) {
        fprintf(stderr, "Error!!!, MultiChannelCamera malloc errr.\n");
        return -1;
    }
    _loop = true;
    _worker_thread = std::thread(&MultiChannelCamera::run, this);
    return 0;
}
int MultiChannelCamera::stop_capturing(void) {
    enum v4l2_buf_type type;
    _loop = false;
    _worker_thread.join();
    if (_fd > 0) {
        type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        if (-1 == xioctl(_fd, VIDIOC_STREAMOFF, &type)) {
            return errno_exit("VIDIOC_STREAMOFF");
        }
    }
    return 0;
}

int MultiChannelCamera::release_device(void) {
    int ret = 0;
    unsigned int i = 0;
    for (i = 0; i < _buffers_count; ++i) {
        if (-1 ==
            release_device_memory(_pbuffers[i].start, _pbuffers[i].length)) {
            fprintf(stderr, "%s release_device_memory ...\\n",
                    _dev_name.c_str());
            ret = -1;
            continue;
        }
    }
    deint_device_memory();
    if (_pbuffers != nullptr) {
        free(_pbuffers);
    }

    return 0;
}

void MultiChannelCamera::close_device(void) {
    if (_fd > 0) {
        close(_fd);
    }
    _fd = -1;
}

int MultiChannelCamera::stop(void) {
    int ret = 0;
    ret = stop_capturing();
    ret |= release_device();
    close_device();
    return ret;
}

void MultiChannelCamera::setConfig(CameraConfig &config) {
    _fps = config.fps;
    _cb_read = config.cb;
    _width = config.width;
    _height = config.height;
    _user_data = config.user_data;
}