
#include "video_processor.h"
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

/*hi3536 用于mac2mac 通信的网口 默认的IP 与 端口 不变*/
#define IP_PORT_DEFAULT 59116
#define IP_ADDR_DEFAULT "192.168.100.251"

/*协议头部*/
#define VIDO_CTL_CMD_HEAD_LOW (0x63)
#define VIDO_CTL_CMD_HEAD_HIGH (0x6C)
#define VIDO_CTL_CMD_HEAD_SIZE (2)

#define PROTOCAL_HEAD_SIZE (sizeof(video_ctrl_cmd_t))
#define PROTOCAL_DATA_MAX_SIZE (1400)

#define PROTOCAL_TIMEOUT_TIME (1500) /*ms*/

/*
 * cmd 返回标识
 */
#define PROTOCAL_ACK_OK "ack <ok>"
#define PROTOCAL_ACK_ERROR "<error>"

#if (VIDEO_PROCESSOR_DEBUG_ENABLE == 1)
#define debug_printf(format, arg...) printf(format, ##arg);
#else
#define debug_printf(format, arg...)                                           \
    {}
#endif

//协议数据，必须按照1字节对齐
#pragma pack(1)
typedef struct video_ctrl_cmd_s {
    uint8_t head_l;
    uint8_t head_h;
    uint16_t len;
    uint32_t index;
    uint8_t data[]; // cmd + payload
} video_ctrl_cmd_t;
#pragma pack()

typedef struct data_cache_buf_s {
    uint16_t len;
    uint8_t buf[PROTOCAL_DATA_MAX_SIZE + PROTOCAL_HEAD_SIZE];
} data_cache_buf_t;

static int s_sk;

static char cmd_tmp_buf[1024];
static uint32_t s_pkt_index = 0;
static struct sockaddr_in send_addr;
static data_cache_buf_t data_cache = {0, {0}};

int _socket_init(const char *ip, uint32_t port) {

    {
        bzero(&send_addr, sizeof(send_addr));
        send_addr.sin_family = AF_INET;
        send_addr.sin_addr.s_addr = inet_addr(ip);
        send_addr.sin_port = htons(port);
    }
    {
        s_sk = socket(AF_INET, SOCK_DGRAM, 0);
        if (s_sk < 0) {
            printf("[VideoProcessor]socket create failure\n");
            return -1;
        }
#if 1
        // set noblock
        int flag = fcntl(s_sk, F_GETFL, 0);
        if (flag < 0) {
            printf("[VideoProcessor]fcntl F_GETFL fail\n");
            return -1;
        }
        if (fcntl(s_sk, F_SETFL, flag | O_NONBLOCK) < 0) {
            printf("[VideoProcessor]fcntl F_SETFL fail\n");
            return -1;
        }
#endif
    }

    {
        struct sockaddr_in server_addr;
        memset(&server_addr, 0, sizeof(struct sockaddr_in));
        server_addr.sin_family = AF_INET;
        server_addr.sin_port = htons(port);
        server_addr.sin_addr.s_addr = htonl(INADDR_ANY); /*any ip*/

        if (bind(s_sk, (struct sockaddr *)&server_addr, sizeof(server_addr)) ==
            -1) {
            printf("[VideoProcessor]binding fail [%s]\n", strerror(errno));
            return -1;
        }
    }
    return 0;
}

void _socket_deinit() { close(s_sk); }

int _recv(char *buf, size_t buf_size, uint32_t pkt_index, uint32_t timeout_ms) {
    struct timeval tv = {0, 0};
    fd_set rd = {0};
    int err = -1;

    tv.tv_sec = timeout_ms / 1000;
    tv.tv_usec = (timeout_ms % 1000) * 1000;

    video_ctrl_cmd_t *p_ctrl_cmd = (video_ctrl_cmd_t *)&data_cache.buf[0];

    memset(data_cache.buf, sizeof(data_cache.buf), 0);
    while (1) {
        FD_ZERO(&rd);
        FD_SET(s_sk, &rd);
        err = select(s_sk + 1, &rd, NULL, NULL, &tv);
        if (err == 0) //超时
        {
            printf("[VideoProcessor]select time out!\n");
            return -2;
        } else if (err == -1) //失败
        {
            printf("[VideoProcessor]fail to select!\n");
            return -1;
        } else //成功
        {
            struct sockaddr_in recv_addr;
            int recv_addr_len = 0;
            int recv_count_cur = 0;
            int recv_count_sum = 0;
            int recv_head_flag = 0;

            if (FD_ISSET(s_sk, &rd)) {
                int need_recv_count = sizeof(data_cache.buf) - recv_count_sum;
                recv_count_cur = recvfrom(
                    s_sk, data_cache.buf + recv_count_sum, need_recv_count, 0,
                    (struct sockaddr *)&recv_addr, (socklen_t *)&recv_addr_len);

                if (recv_head_flag == 0) {
                    if (recv_count_cur + recv_count_sum >=
                        VIDO_CTL_CMD_HEAD_SIZE) {
                        for (size_t i = 0;
                            i < recv_count_cur + recv_count_sum - 1; i++) {
                            if ((data_cache.buf[i] == VIDO_CTL_CMD_HEAD_LOW) ||
                                (data_cache.buf[i + 1] ==
                                 VIDO_CTL_CMD_HEAD_LOW)) {
                                recv_head_flag = 1;
                                recv_count_cur = recv_count_cur - i;
                                break;
                            }
                        }
                        if (data_cache.buf[recv_count_cur + recv_count_sum -
                                           1] == VIDO_CTL_CMD_HEAD_LOW) {
                            recv_count_sum = 1;
                        }
                    }
                }
                if (recv_head_flag == 1) {
                    recv_count_sum += recv_count_cur;
                    if (recv_count_sum >= sizeof(video_ctrl_cmd_t)) {
                        if (recv_count_sum >= ntohs(p_ctrl_cmd->len)) {
                            // Recv one pkt
                            break;
                        }
                    }
                }
            } else {
                printf("[VideoProcessor]It's not supposed to run here.\n");
                return -3;
            }
        }
    }

    debug_printf("[VideoProcessor]Recv pkt len[%d] index[%x] data[%s]\n",
                 ntohs(p_ctrl_cmd->len), ntohl(p_ctrl_cmd->index),
                 p_ctrl_cmd->data);

    if (pkt_index != ntohl(p_ctrl_cmd->index)) {
        printf("[VideoProcessor]Error Index send[%d] recv[%d] [%08x]\n",
               pkt_index, ntohl(p_ctrl_cmd->index), p_ctrl_cmd->index);
        return -2;
    }
    int data_len = p_ctrl_cmd->len - sizeof(video_ctrl_cmd_t);
    uint32_t min = data_len;
    if (buf != NULL) {
        min = (data_len > buf_size) ? buf_size : data_len;
        memcpy(buf, p_ctrl_cmd->data, min);
    }
    return min;
}

int _send(const char *cmd, const char *data) {
    video_ctrl_cmd_t *p_ctrl_cmd = (video_ctrl_cmd_t *)&data_cache.buf[0];

    if ((cmd == NULL) || (data == NULL)) {
        printf("[VideoProcessor]Param is error.\n");
        return -1;
    }
    uint16_t pkt_len = sizeof(video_ctrl_cmd_t) + strlen(cmd) + strlen(data);
    s_pkt_index++;
    p_ctrl_cmd->head_l = VIDO_CTL_CMD_HEAD_LOW;
    p_ctrl_cmd->head_h = VIDO_CTL_CMD_HEAD_HIGH;
    p_ctrl_cmd->index = ntohl(s_pkt_index); // 大端
    p_ctrl_cmd->len = ntohs(pkt_len);       // 大端

    if (pkt_len > PROTOCAL_DATA_MAX_SIZE) {
        printf("[VideoProcessor]Send cmd data too long [%d]-[%d]\n", pkt_len,
               PROTOCAL_DATA_MAX_SIZE);
        return -1;
    }

    strcpy((char *)p_ctrl_cmd->data, cmd);
    strcpy((char *)p_ctrl_cmd->data + strlen(cmd), data);

#if (VIDEO_PROCESSOR_DEBUG_ENABLE == 1)
    {

        printf("\n\n[Debug]: Send Cmd --> len[%d] --> data[\n\t", pkt_len);
        for (int i = 0; i < pkt_len; i++) {
            printf("%02x ", data_cache.buf[i]);
            if ((i != 0) && (i % 15 == 0)) {
                printf("\n\t");
            }
        }
        printf("]\n\n");
    }
#endif

    size_t len = sendto(s_sk, data_cache.buf, pkt_len, 0,
                        (struct sockaddr *)&send_addr, sizeof(send_addr));
    if (len != pkt_len) {
        printf("[VideoProcessor]Send cmd data failed. errro[%s] send_len[%d] "
               "pkt_len[%d] s_sk[%d]\n",
               strerror(errno), len, pkt_len, s_sk);
        return -1;
    }
    return len;
}

int _sendSetCmd(const char *cmd, const char *data) {
    int ret = 0;
    char recv_buf[10];
    if (data == NULL) {
        printf("[VideoProcessor]Param is error.\n");
        return -1;
    }

    debug_printf("\nSend cmd-->[%s] data-->[%s] .\n", cmd, data);
    ret = _send(cmd, data);
    if (ret < 0) {
        printf("[VideoProcessor]Send Cmd :[%s] [%s]is error.\n", cmd, data);
        return -1;
    }

    ret = _recv(recv_buf, sizeof(recv_buf), s_pkt_index, PROTOCAL_TIMEOUT_TIME);
    if (ret <= 0) {
        printf("[VideoProcessor]Recv Error ==> Send Cmd :[%s] [%s]is error.\n",
               cmd, data);
        return -1;
    }
#if (VIDEO_PROCESSOR_DEBUG_ENABLE == 1)
    {
        printf("\n\n[Debug]: Recv ACK --> len[%d] --> data[", ret);
        for (int i = 0; i < ret; i++) {
            printf("%c", recv_buf[i]);
        }
        printf("]\n\n");
    }
#endif

    if (memcmp(PROTOCAL_ACK_OK, recv_buf, strlen(PROTOCAL_ACK_OK)) != 0) {
        printf("[VideoProcessor] Recv Error ACK ==> Send Cmd :[%s] [%s]is "
               "error.\n",
               cmd, data);
        return -1;
    }
    debug_printf("Send [%s][%s] Success.\n\n", cmd, data);
    return 0;
}

/*
* Function: 设置Video Processor 网络参数

* mac:      23:50:95:da:65:fe
* ip:       192.168.1.125
* netmask:  255.255.255.0
* gw:       192.168.1.1
*/

int video_processor_set_network(NetWorkConfigParam_t *network_cfg) {
    int ret = 0;
    if ((network_cfg->mac == NULL) || (network_cfg->ip == NULL) ||
        (network_cfg->netmask == NULL) || (network_cfg->gw == NULL)) {
        printf("[VideoProcessor] VideoProcessorSetNetWork param is null.\n");
        return -1;
    }

    ret = snprintf(cmd_tmp_buf, sizeof(cmd_tmp_buf),
                   "<mac:%s,ip:%s,netmask:%s,gw:%s>", network_cfg->mac,
                   network_cfg->ip, network_cfg->netmask, network_cfg->gw);
    if (ret > sizeof(cmd_tmp_buf)) {
        printf("[VideoProcessor] VideoProcessorSetNetWork cmd too long "
               "[%d],only support[%d]\n",
               ret, sizeof(cmd_tmp_buf));
        return -1;
    }
    return _sendSetCmd("setN", cmd_tmp_buf);
}

/*
 * Function: 设置通道 RTSP流地址
 * ch： 范围 0-15
 * rtsp: "rtsp://admin:admin123456@192.168.1.64:554//Streaming/Channels/1>"
 */
int video_processor_set_source(uint16_t ch, const char *rtsp) {
    int ret = 0;
    if (ch > 15) {
        printf("[VideoProcessor] VideoProcessorSetSource support ch 0-15,not "
               "support [%d].\n",
               ch);
        return -1;
    }
    if ((rtsp == NULL)) {
        printf("[VideoProcessor] VideoProcessorSetSource rtsp is null.\n");
        return -1;
    }
    ret =
        snprintf(cmd_tmp_buf, sizeof(cmd_tmp_buf), "<win-1%02d:%s>", ch, rtsp);
    if (ret > sizeof(cmd_tmp_buf)) {
        printf("[VideoProcessor] VideoProcessorSetSource cmd too long "
               "[%d],only support[%d]\n",
               ret, sizeof(cmd_tmp_buf));
        return -1;
    }
    return _sendSetCmd("setP", cmd_tmp_buf);
}

/*
 * Function:设置帧率
 * frame_rate: 目前只支持 P24 与 P30
 */
int video_processor_set_frame_rate(VideoProcessorFrameRate_t frame_rate) {
    int ret = 0;

    if (frame_rate == VIDEO_PROCESSOR_FRAMERATE_P24) {
        ret = _sendSetCmd("setO", "<out-0:1920x1080@24>");
    } else if (frame_rate == VIDEO_PROCESSOR_FRAMERATE_P30) {
        ret = _sendSetCmd("setO", "<out-0:1920x1080@30>");
    } else {
        ret = -1;
    }

    return ret;
}

/*
 * Function: 设置通道时序
 * seq: <ch1:2,ch2:3,ch15:2,ch4:1>
 * 表示：ch1,ch1,ch2,ch2,ch2,ch15,ch15,ch4 ..... 重复这个顺序
 */
int video_processor_set_channel_sequence(const char *data) {
    return _sendSetCmd("setC", data);
}

int video_processor_init(void) {
    return _socket_init(IP_ADDR_DEFAULT, IP_PORT_DEFAULT);
}
void video_processor_deinit(void) {
    _socket_deinit(); 
}